in vec2 attrib0;
in vec2 attrib1;
in vec4 attrib2; // pos.xy, size, depth
in vec4 attrib3; // tex offset, alpha, flip

uniform vec4 resolution;

out vec4 params0;
out vec4 coord0;

void main() {
    vec2 pos = attrib0 * vec2(attrib3.z,resolution.z * 0.5) * attrib2.z + attrib2.xy;

    coord0.xy = attrib1;
    coord0.y = coord0.y * 0.25 + attrib3.x;
    coord0.zw = pos;
    params0 = vec4(attrib2.w, attrib3.zxy);

    gl_Position = vec4(pos, 1.0, 1.0);
}

[FS]
#ifdef GL_FRAGMENT_PRECISION_HIGH
    precision highp float;
#else
    precision mediump float;
#endif
precision lowp sampler2D;

uniform lowp sampler2D tex0; // data 1 (dense)
uniform lowp sampler2D tex1; // data 2 (dense)
#ifdef DENSITY_LERP
    uniform lowp sampler2D tex2; // data 1
    uniform lowp sampler2D tex3; // data 2
#endif

uniform lowp vec4 sun;
uniform lowp vec4 sun_color; // linear
uniform lowp vec4 ambient;
uniform lowp vec4 fog_params;
uniform vec4 resolution;

in vec4 params0;
in vec4 coord0;

#if __VERSION__ >= 300
    layout(location = 0) out vec4 fragData_0;
#endif

const float PI = 3.14159265;
const float DEG2RAD = PI / 180.0;
const float FOV = 90.0;
const float FOCAL_LENGTH = 1.0 / tan(FOV * DEG2RAD * 0.5);
const float AMB_ENEGRY = 1.5;
const float LIGHT_ENEGRY = 1.5;

// gamma correction
float toLinear(float c) { return c * c; }
vec3 toLinear(vec3 c) { return c * c; }
vec3 toSRGB(vec3 c) { return sqrt(c); }

// directions of the precomputed light
const vec3 dir00 = normalize(vec3(1,-1,1));
const vec3 dir01 = normalize(vec3(-1,-1,1));
const vec3 dir02 = normalize(vec3(-1,1,1));
const vec3 dir03 = normalize(vec3(1,1,1));
const vec3 dir10 = normalize(vec3(0,-1,0));
const vec3 dir11 = normalize(vec3(0,0,1));

vec3 getDirection(vec2 uv_ss) {
    vec2 uv_vs = uv_ss * vec2(1.0,-resolution.w);
    return normalize(vec3(uv_vs, FOCAL_LENGTH));
}

lowp float getFogDensity(float distanceM) {
    return 1.0 - exp(-fog_params.w * distanceM);
}

float integrateSunLight(vec2 pos, in vec4 data0, in vec4 data1, float s) {
    vec3 L = getDirection((pos - coord0.zw) * 1.0);
    //vec3 V = getDirection(coord0.zw);
    L.x *= params0.y;

    float luminance = 0.0;
    float total_w = 0.0, w;

#define LIGHT(lum,dir) \
    w = max(dot(dir,L),0.0); w *= w * w; \
    luminance += w * lum; \
    total_w += w;

    LIGHT(data0.x,dir00)
    LIGHT(data0.y,dir01)
    LIGHT(data0.z,dir02)
    LIGHT(data0.w,dir03)
    LIGHT(data1.x,dir10)
    LIGHT(data1.y,dir11)
#undef LIGHT

    float phase = pow(max(dot(L,vec3(0.0,0.0,-1.0)),0.0), 2.0 * s) * 2.0 * s + 0.5;
    return luminance / total_w * LIGHT_ENEGRY * phase;
}

// main
void main() {
#ifdef DENSITY_LERP
    vec4 data0 = mix(texture(tex2,coord0.xy), texture(tex0,coord0.xy), params0.z);
    vec4 data1 = mix(texture(tex3,coord0.xy), texture(tex1,coord0.xy), params0.z);
#else
    vec4 data0 = texture(tex0,coord0.xy);
    vec4 data1 = texture(tex1,coord0.xy);
#endif

    // ambient
    float total_lum = (0.25 + 0.75 * data1.z) * AMB_ENEGRY; // ambient

    // lighting
    total_lum += sun_color.y * integrateSunLight(sun.xy,data0,data1,1.0); // sun

    vec3 col = mix(toLinear(vec3(0.57,0.61,0.65)), toLinear(vec3(0.91,0.89,0.87)), total_lum); // ambient
    col *= ambient.xyz;

    col += sun_color.x * integrateSunLight(sun.zw,data0,data1,2.0); // moon

    // fog
    col = mix(col, fog_params.xyz, getFogDensity(params0.x));

    gl_FragColor = vec4(toSRGB(col), data1.w * params0.w);
}

