attribute vec2 attrib0;
attribute vec2 attrib1;
uniform lowp vec4 params;
uniform lowp vec4 params2;
varying vec2 coord0;
varying lowp vec2 proj;

const float PI = 3.141592;

void main() {
    coord0 = attrib1;
    float ang_x_rad = params.y / 180.0 * PI;
    proj.x = params2.x * tan(-ang_x_rad);

    gl_Position = vec4(attrib0, 1.0, 1.0);
}

[FS]
#ifdef GL_FRAGMENT_PRECISION_HIGH
    precision highp float;
#else
    precision mediump float;
#endif
precision lowp sampler2D;

uniform lowp sampler2D tex0; // reflection
uniform sampler2D tex1; // mask
uniform sampler2D tex2; // ripple
uniform sampler2D tex3; // geometric wave
uniform sampler2D tex4; // foam
#ifdef INTERACTIVE
    uniform sampler2D tex5; // interactive rain
    uniform sampler2D tex6; // interactive touch
#endif
uniform lowp sampler2D tex7; // photo

uniform lowp vec4 params; // camY, angX, wind speed norm, biasY
uniform lowp vec4 params2; // dirZ, fresnelF0, rippleHeight, reflectivity
uniform lowp vec4 params3; // lengthy reflections, smooth, LF wave height, LF wave freq
uniform lowp vec4 params4; // LF wave choppiness, reflection percent
uniform lowp vec4 params5; // foam decay, foam threshold
uniform lowp vec4 params6; // rain scale, touch scale
uniform lowp vec4 color;
uniform lowp vec4 daytime_color;
uniform vec4 offset;
uniform lowp vec4 fog_params; // color, mistK * distance scale
uniform lowp vec4 fog_params2; // units2meters, 
uniform lowp vec4 sun_params;
uniform lowp vec4 sun_color;
uniform lowp vec4 ambient;
uniform lowp mat2 rotation;
uniform vec2 time;
uniform lowp vec4 touch;
uniform lowp vec2 uv_ratio;
uniform vec4 resolution;

varying vec2 coord0;
varying lowp vec2 proj;

#if __VERSION__ >= 300
    layout(location = 0) out vec4 fragData_0;
#endif

// #define SETUP_MODE
#define FLIP_Y
#define WIND_SPEED params.z
// #define SEA_TRACING_MODE 1

const int NUM_STEPS = 8;
const float HORIZONTAL_STRETCH = 0.5;
const vec2 INTERACTIVE_TOUCH_OFFSET = vec2(0.5,0.75);


vec3 toLinear(vec3 c) { return c * c; }
vec3 toSRGB(vec3 c) { return sqrt(c); }

vec3 intersectPlaneY(vec3 o, vec3 d) {
    return o - d * (o.y / d.y);
}
vec2 scaleAround(vec2 x, vec2 p, float s) { return (x - p) * s + p; }

vec2 reflectPointByPlane(vec2 p, vec3 plane) {
    lowp float dist = dot(plane.xy,p) + plane.z;
    return p - plane.xy * dist * 2.0;
}

float intersectPlaneRay(vec3 plane, vec2 ori, vec2 dir) {
    return (dot(plane.xy,ori)+plane.z) / dot(plane.xy,-dir);
}

vec3 combineNormals1(vec3 n0, vec3 n1) {
    return vec3(n0.xy + n1.xy, n0.z * n1.z);
}

float getLFWaveHeightMultiscale(vec3 p, out vec3 n) {
    lowp vec2 uv = vec2(p.xz * params3.w * 0.5);
    mediump vec4 data = texture(tex3,uv) * 2.0 - 1.0;
    n = data.xyz;
    return 2.0 * p.y - data.w * params3.z;
}

float heightMapTracing(inout highp vec3 ori, vec3 dir, float wave_height, out vec3 p, out highp vec3 n) {  
    float choppyM = 6.0 / float(NUM_STEPS*3) * params4.x / (128.0 * params3.w); 
    mediump float tmid = 0.0;

#if SEA_TRACING_MODE == 1
    float d = 0.0;
    for(lowp int i = 0; i < NUM_STEPS; i++) {
        p = ori + dir * tmid;
    	d = getLFWaveHeightMultiscale(p, n);
        ori.xz += n.xy * choppyM; // choppy
        tmid -= d * 0.4;
    } 
    for(lowp int i = 0; i < NUM_STEPS; i++) {
        p = ori + dir * tmid;
    	d = getLFWaveHeightMultiscale(p, n);
        ori.xz += n.xy * choppyM; // choppy
        tmid -= d * 0.5;
    }
#endif
   
    mediump float tm = tmid;
    mediump float tx = 1000.0;
    float hx = getLFWaveHeightMultiscale(ori + dir * tx, n);
    float hm = getLFWaveHeightMultiscale(ori + dir * tm, n);  
    for(lowp int i = 0; i < NUM_STEPS; i++) {
        tmid = mix(tm, tx, hm / (hm - hx));                   
        p = ori + dir * tmid;
    	float hmid = getLFWaveHeightMultiscale(p, n);
        ori.xz += n.xy * choppyM; // choppy
		if(hmid < 0.0) {
        	tx = tmid;
            hx = hmid;
        } else {
            tm = tmid;
            hm = hmid;
        }
    }
    n = normalize(n);
    return tmid;
}

float traceWaterSurface(inout highp vec3 o, in vec3 d, float wave_height, out vec3 p, out mediump vec3 n, out float dist) {
    dist = heightMapTracing(o,d,wave_height,p,n);
    
    // ripple normal
    vec2 rns = vec2(params2.z,1.0);
    vec2 nuv = p.xz * params4.zw + offset.xy;
    lowp vec3 rn = normalize((texture(tex2, nuv           + vec2(time.y,time.y*3.)).xyz * 2.0 - 1.0) * rns.xxy);
    rn = combineNormals1(rn, normalize((texture(tex2, nuv*(1.0/1.5) - vec2(time.y,time.y*5.)).xyz * 2.0 - 1.0) * rns.xxy));
    rn = combineNormals1(rn, normalize((texture(tex2, nuv*(1.0/1.35) + vec2(time.y*2.,time.y)).xyz * 2.0 - 1.0) * rns.xxy));
    rn.y = -rn.y;

    // combine
    n = combineNormals1(n,rn);

    // interactive
    #ifdef INTERACTIVE
        vec3 interactive_normal = texture(tex5, p.xz * params6.x).xyz * 2.0 - 1.0;
        vec3 interactive_touch_normal = texture(tex6, p.xz * params6.y + vec2(0.5,touch.w)).xyz * 2.0 - 1.0;
        interactive_normal.xy *= (1.0 + WIND_SPEED * 3.0);
        interactive_touch_normal.xy *= (1.0 + WIND_SPEED * 3.0);
        n = combineNormals1(n,normalize(interactive_normal));
        n = combineNormals1(n,normalize(interactive_touch_normal));
    #endif

    return 1.0;
}

float getSunLight() {
    mediump float sun_light = distance(coord0.x * resolution.x, sun_params.x);
    sun_light = smoothstep(fog_params2.y*3.0+sun_params.z*0.5, fog_params2.y-sun_params.z*0.7, sun_light); // section length instead of radius
    sun_light *= sun_light;
    
    lowp float vertical_falloff = max(1.0 - distance(coord0.y * resolution.y, fog_params2.z) / (sun_params.z * 0.5), 0.0);
    sun_light *= vertical_falloff * vertical_falloff;

    mediump float horizon_proximity = distance(coord0.y * resolution.y, sun_params.y);
    horizon_proximity = smoothstep(sun_params.z*1.5, sun_params.z * 0.5, horizon_proximity);

    return sun_light * horizon_proximity;
}

lowp float getFogDensity(float distanceUnits, mediump float sun_light) {
    // distanceUnits = min(distanceUnits, 250.0);
    lowp float distanceM = max(distanceUnits * fog_params2.x, 0.0);
    return max(1.0 - exp(-fog_params.w * distanceM), 0.0) * (1.0 - sun_light * fog_params2.w * sun_color.w);
}

vec3 getReflection(vec2 uv, float lod) {
    mediump float sun_light = distance(uv * resolution.xy, sun_params.xy);
    sun_light = smoothstep(sun_params.z*2.0, sun_params.z*0.9, sun_light);
    sun_light *= sun_light * sun_params.w;
    uv.y -= 0.00125;
    return texture(tex0, uv).xyz + sun_color.xyz * sun_color.w * sun_light * ambient.w * 0.4;
}

const float FOAM_UV_SCALE = 0.5;
float getFoam(vec2 uv, float zdist, float smoothness, float turbulence) {
    vec2 foam_uv = uv * FOAM_UV_SCALE + offset.zw * 4.0;
    float foam = texture(tex4, foam_uv).x;
    float foam_threshold = zdist * params5.y + params5.x;
    return smoothstep(foam_threshold,foam_threshold+smoothness,foam);
}

float getGlobalFoam(vec2 uv, float smoothness, float turbulence) {
    vec2 foam_uv = uv * FOAM_UV_SCALE * 0.7 + offset.zw * 4.0;
    float foam = texture(tex4, foam_uv.yx).x;
    float foam_threshold = 1.0 - turbulence;
    return smoothstep(foam_threshold,foam_threshold+smoothness,foam);
}

vec2 toMaskSpace(vec2 uv) {
    uv = scaleAround(uv,vec2(0.5,0.0),touch.z);
    uv += touch.xy;
    uv.y /= uv_ratio.y;
    uv = uv * 2.0 - 1.0;
#ifdef FLIP_Y
    uv.y = -uv.y;
#endif
    uv.x *= resolution.z * HORIZONTAL_STRETCH; 
    return uv;
}

vec2 toImageSpace(vec2 uv) {
    uv.x *= resolution.w / HORIZONTAL_STRETCH; 
#ifdef FLIP_Y
    uv.y = -uv.y;
#endif
    uv = uv * 0.5 + 0.5;
    uv.y *= uv_ratio.y;
    uv -= touch.xy;
    uv = scaleAround(uv,vec2(0.5,0.0), 1.0/touch.z);
    return uv;
}

vec3 getSunDirection() {
    vec2 mask_uv = toMaskSpace(sun_params.xy / resolution.xy);
    mask_uv.y = mask_uv.y * 0.5 + 0.5;
    vec3 dir = vec3(mask_uv, params2.x); 
    return normalize(dir);
}

float blinn(vec3 light, vec3 eye, vec3 normal, float shiness){
	float nrm = (shiness + 8.0) / (3.1415 * 8.0);
	vec3 _half = normalize(light+eye);
	return pow(max(dot(_half,normal),0.0),shiness) * nrm;
}

// main
void main() {
    lowp vec2 uv = toMaskSpace(coord0);
#ifndef NO_MASK
    vec2 mask_uv = uv;
    mask_uv.x *= resolution.w / HORIZONTAL_STRETCH; 
    lowp float mask_tex = texture(tex1, mask_uv*0.5+0.5).x;
    if(mask_tex < 0.001) discard;
#endif  

    // camera
    lowp vec3 dir = vec3(uv, params2.x);
    dir.yz *= rotation;
    highp vec3 ori = vec3(0.0,params.x,3.0);
    highp vec3 ori_choppy = ori;
    dir = normalize(dir);  

    // flat reflection plane
    lowp vec3 refl_plane = vec3(0.0,1.0,-proj.x);
    lowp float dist = uv.y - proj.x;
    lowp float wave_height = 1.0;

    // trace the water
    vec3 p;
    mediump vec3 n;
    lowp vec3 wcol;
    float p_dist;
    lowp float trace_mask = traceWaterSurface(ori_choppy,dir,wave_height,p,n,p_dist);    

    // normal in screen space
    lowp vec2 nss = normalize(n.xz * vec2(0.25,1.0));
    
    // reflection
    dist = (uv.y - proj.x) / nss.y; // intersectPlaneRay
    lowp float lod = log2(1.0+dist*params3.y); // blur stars & avoid derivatives pixellation tradeoff
    lowp vec2 refl_offset = nss * (uv.y - proj.x) * 2.0;

    // flip along horizon line
    lowp float refl_ray_scale = max(params3.x * (n.y + 1.0), 0.5);
    lowp float horizon_offset = refl_offset.y * 0.5;
    refl_offset *= refl_ray_scale;
    refl_offset.y -= (refl_offset.y - horizon_offset) * 2.0;

    lowp vec2 ruv = uv - refl_offset;
    lowp vec2 ruv_is = toImageSpace(ruv);
    lowp vec3 reflected = getReflection(ruv_is, lod);
    reflected *= reflected;
    // reflected *= lod + 1.0; // simulate HDR for night

    // refraction
    lowp float hh = clamp(1.0 + p.y, 0.0,1.0);    
    lowp vec3 refracted = color.xyz * hh;
    float sun_light = getSunLight();

#ifdef USE_REAL_REFLECTION
    lowp vec2 rfr_uv = uv - (nss) * params3.x * n.y * 0.5;
    rfr_uv.x *= resolution.w / HORIZONTAL_STRETCH;

    lowp float mask_tex_refl = texture(tex1, rfr_uv*0.5+0.5).x;
    
    // skip distortion if it points outside the mask
    if(mask_tex_refl < 1.0) rfr_uv = uv * vec2(resolution.w / HORIZONTAL_STRETCH, 1.0);
    else rfr_uv.y = max(rfr_uv.y, proj.x * 0.99);

    lowp vec3 refracted_image = texture(tex7, rfr_uv*0.5+0.5).xyz * daytime_color.xyz;
    refracted_image = toLinear(refracted_image);

    float traced_reflection_around_sun = max(1.0-sun_light * 2.0, 0.0);
    float original_alpha = 1.0 - (1.0-params4.y) * traced_reflection_around_sun;
    reflected = mix(refracted_image, reflected, original_alpha);
    refracted = mix(refracted_image, refracted, original_alpha);
#endif
    
    float fog_density = getFogDensity(distance(p, ori), sun_light);
    float y_alpha = clamp(smoothstep(-1.0,-0.99,uv.y), 0.0,1.0);

    // SSS
#ifdef SSS
    vec3 L = getSunDirection() * vec3(1.,1.,1.);
    const float SUBSRF_FACTOR = 4.0;
    float sss = SUBSRF_FACTOR * sun_params.w *
        pow(clamp(p.y+1.5,0.0,1.0),1.0) * // height
        pow(max(dot(dir,L),0.0),3.0) * // sun direction
        pow(max(dot(n.xzy,L)*0.7+0.3,0.0), 2.0); // surface normal
    refracted += color.xyz * toLinear(sun_color.xyz) * sss;
#endif

    float zdist = ori.z - p.z;
    vec2 flat_p = ori.xz - dir.xz * (ori.y / dir.y);    
    vec2 foam_uv = mix(p.xz, flat_p, params5.w);

#ifdef OCEAN
    #ifdef DEEP_WATER
        float deep_water_alpha = sqrt(smoothstep(0.0,50.0,zdist)) * (1.0 - sun_light) * 0.6 * y_alpha;
        vec3 deep_water = mix(vec3(1.0), vec3(0.0,0.2,0.8), deep_water_alpha);
        refracted *= deep_water;
        reflected *= deep_water;
    #endif

    #ifdef BEACH
        vec3 SAND = vec3(0.869,0.689,0.523) * ambient.xyz;
        refracted = mix(SAND * SAND * SAND, refracted, pow(smoothstep(0.0,2.0,zdist), 0.6));
    #endif
#endif

    float turbulence = clamp((p.y / max(params3.z,0.1) * 0.5 + 0.5) * (1.9 - params5.x * 1.5), 0.0,1.0);
    turbulence = min(turbulence, 1.9 - params5.x * 1.9);
    turbulence = turbulence * turbulence * 0.8;    

    #if defined(FOAM_SHADOW) && defined(FOAM_LOCAL) && defined(BEACH)
        vec3 flat_choppy_p = ori_choppy + dir * p_dist * max(1.0, (0.6 + zdist * 0.8));
        refracted *= 1.0 - getFoam(flat_choppy_p.xz,zdist,1.0, turbulence);
    #endif

    // mix
    vec3 fwd_dir = normalize(vec3(0.0, dir.yz));
    float fade = smoothstep(0.7,0.5,abs(ruv_is.y-0.5));
    float fresnel = clamp(pow(abs(dot(n.xzy,fwd_dir)),1.0/5.0), 0.0,1.0);
    fresnel = params2.y + fresnel * (1.0 - params2.y) * params2.w;
    fresnel = 1.0 - (1.0 - fresnel) * fade;
    fresnel *= y_alpha;
    wcol = mix(reflected, refracted, fresnel);
    wcol = sqrt(wcol);    

    // foam
#if defined(FOAM_LOCAL) || defined(FOAM_GLOBAL)
    float foam_level = 1.0;
    #ifdef FOAM_LOCAL
        foam_level *= (1.0 - getFoam(foam_uv,zdist,params5.z,turbulence) * 0.8);
    #endif
    #ifdef FOAM_GLOBAL
        foam_level *= (1.0 - getGlobalFoam(foam_uv,params5.z,turbulence) * 0.6);
    #endif
    wcol = mix(ambient.xyz, wcol, foam_level);
#endif

    // fog
    wcol = mix(wcol, fog_params.xyz, fog_density);

    // debug
    // wcol = vec3(fog_density);
    // wcol = n * 0.5 + 0.5;
    // wcol = texture(tex5,p.xz).xyz;
    // vec3 nn = texture(tex6,coord0).xyz * 2.0 - 1.0;
    // nn.xy *= 50.0;
    // wcol = normalize(nn) * 0.5 + 0.5;
    // wcol = texture(tex0,coord0*2.0).xyz;
 
    // combine
    lowp float mask = step(0.0,dir.y) * color.w;
#ifndef NO_MASK
    mask *= mask_tex;
#endif

    gl_FragColor = vec4(wcol, mask);
}

